#ifndef _GRDB_H_
#define _GRDB_H_

#include <set>
#include <vector>
#include <unordered_map>
#include <cassert>
#include <cstring>
#include <cmath>
#include <iostream>
#include <fstream>
#include <stdexcept>
#include <algorithm>

#include "baseDB.h"
#include "layoutDB.h"

class RoutingDB
{

        //
        //  Layout Information
        //
    public:
        RoutingDB( const string & designName, const Layout & layout ) : _designName( designName ), _numTilesX( layout._numTilesX ), _numTilesY( layout._numTilesY ), 
                                                                        _numLayers( layout._numLayers ), _numGlobalNets(0), _tileWidth( static_cast<int>(layout._cellWidth) ), 
                                                                        _tileHeight( static_cast<int>(layout._cellHeight) ) { }
    private:
        string _designName;
        int _numTilesX, _numTilesY, _numLayers;
        vector<RoutingDir> _dirLayers;
        vector<int> _trackDemands;
        vector<int> _cntEdgeLayers;
        vector<int> _capOnLayer;
        vector<int> _upViaSize;
        vector<int> _dnViaSize;
        int _tileWidth, _tileHeight;

        //
        //  Edge Related Objects
        //
    public:
        void initEdges( const Layout & layout );
        void initGlobalEdgeProfile( const Layout & layout );
    private:
        void refreshNodeBlkInfo( const Layout & layout, const Node & node );
        void refreshRectBlkInfo( const Layout & layout, const Shape & shape, const int layerId );
        void updateEdgeFreeList( const Layout & layout, const int edgeId, const Shape & shape );
        void updateSingleEdgeFreeList( Edge & edge, const double blkStart, const double blkEnd );

        int findEdge( const int x, const int y, const int z ) const;
        Gcell getEdgeLoc( const int edgeId ) const;
        int getEdgeLayer( const int edgeId ) const;
        int getEdgeDemand( const int edgeId ) const;
        int getEdgeBlk( const int edgeId ) const;
        int getEdgeCap( const int edgeId ) const;
        int getEdgeOf( const int edgeId ) const;
        void addEdgeDemand( const int edgeId );

    public:
        void clearEdgeDemands();
    private:

        void checkEdgeFreeList( const int edgeId );
        //  gWire related methods
        bool isUpViaGlobalWire( const Gnet & gnet, const GlobalWire & gWire ) const;
        bool isDnViaGlobalWire( const Gnet & gnet, const GlobalWire & gWire ) const;
        void bruteForceFindAllChildWires( const Gnet & gnet, const int wireId, vector<int> & cache );

        inline void rmEdgeDemand( const int edgeId );
        vector<Edge> _edges;
        void printEdgeUsage( const size_t edgeId ); 
        void printCompleteEdgeInfo( const int edgeId );


        //  
        //  Via Related Objects
        //
    public:
        void initVias( const Layout & layout );
    private:
        long findVia( const int x, const int y, const int z ) const;
        void addViaDemand( const int viaId );
        vector<Via> _vias;

        //
        //  Read Global Routes
        //
    public:
        void initGlobalNets( Layout & layout );
        void readGlobalWires( const Layout & layout, const char * wiresFile );
        void readGlobalWires( const Layout & layout, const string & wiresFile );
        void initRoutingTreeForAllGnets();
        bool checkAllRoutingTrees();
    private:
        void initRoutingTree( Gnet & gnet );

        //
        //  Routing Tree objects
        //  
        vector<vector<PtInfo> > _treeTable;
        // Structure of PtInfo is in baseDB.h: (int x, int y, int z, int indexId, int distPrev, int distNext, bool hasDnVia, bool hasUpVia)
        vector<PtInfo> tmp;

        vector<unordered_map<Gcell, int, HashGcell2d> > _uMapCell2LocInTable;
        vector<Gcell> frontier;
        vector<Gnet> _gnets;
        int _numGlobalNets;
        /* used for removing antenna wires */
        vector<int> _offset;   
        vector<int> _numChildren;
        vector<bool> _validWires;
        vector<GlobalWire> _cleanWires;

    public:
        void updateEdgeDemands();

    private:
        void createBlockage( const int xStart, const int yStart, const int xEnd, const int yEnd, const int z, Layout & layout );


};

#endif
